
#if defined(NUM_RENDER_THREADS) || defined(NUM_BUILD_THREADS)
#  include <pthread.h>
#endif
#include "XRay.hxx"
#include <stdio.h>
#include <stack>
#include "base/Clock.hxx"

#ifdef USE_NUMA
#include <numa.h>
#include <sched.h>
#else
int numa_max_node() { return 0; }
#endif

#ifdef NUM_BUILD_THREADS
void *preprocessThread(void *rayTraceRenderer);
pthread_t build_thread[NUM_BUILD_THREADS];
#endif
int volatile preprocess_threadIDsCounter = 0;

static pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER;

RayTraceRenderer::RayTraceRenderer()
{
#ifdef NUM_BUILD_THREADS
  for (int i = 0; i < NUM_BUILD_THREADS; i++)
    pthread_create(&build_thread[i], NULL, preprocessThread, this);
  while (preprocess_threadIDsCounter < NUM_BUILD_THREADS) { sleep(0); }
  pthread_mutex_lock( &mutex );
  pthread_mutex_unlock( &mutex );
#endif

#ifdef USE_NUMA
  if (numa_available() < 0) {
    printf("Your system does not support NUMA API\n");
  }
#endif
}

int RayTraceRenderer::getBuildPreprocessNode(int myThreadID)
{
#if defined(__linux__)
  /* Determine the actual number of processors */
  const int NUM_PROCS = sysconf(_SC_NPROCESSORS_CONF);

  const int CPU_PER_NODE = NUM_PROCS / (numa_max_node()+1);
  return getBuildPreprocessCPU(myThreadID) / CPU_PER_NODE;
#else
  return 0;
#endif //__linux__
}

int RayTraceRenderer::getBuildPreprocessCPU(int myThreadID)
{
#if defined(__linux__)
  /* Determine the actual number of processors */
  const int NUM_PROCS = sysconf(_SC_NPROCESSORS_CONF);

  const int CPU_PER_NODE = NUM_PROCS / (numa_max_node()+1);
  
  //spread out threads among nodes and then cpus.
  const int cpu = static_cast<int>(myThreadID * CPU_PER_NODE + 
                    myThreadID / (float)(numa_max_node()+1)
                  ) % NUM_PROCS;
  return cpu;
#else 
  return 0;
#endif //__linux__
}

#if defined(NUM_RENDER_THREADS) || defined(NUM_BUILD_THREADS) || \
    defined(NUM_TRI_INTERSECT_THREADS)
static pthread_cond_t condition = PTHREAD_COND_INITIALIZER;
static pthread_cond_t condition_start = PTHREAD_COND_INITIALIZER;
static pthread_cond_t condition_preprocess_start = PTHREAD_COND_INITIALIZER;
static pthread_cond_t condition_end = PTHREAD_COND_INITIALIZER;

int volatile runningCounter = -1;

void *preprocessThread(void *rayTraceRenderer) {
    pthread_mutex_lock( &mutex );
    int myThreadID = preprocess_threadIDsCounter++;

#ifdef __linux__
    const int cpu = RayTraceRenderer::getBuildPreprocessCPU(myThreadID);
    cout << "acceleration structure preprocess on "<< cpu<<endl;

	cpu_set_t mask;
    CPU_ZERO( &mask );
    CPU_SET( cpu, &mask );

    sched_setaffinity(0, sizeof(mask), &mask);
#endif //__linux__

    while(true) {
        pthread_cond_wait( &condition_preprocess_start, &mutex );
        pthread_mutex_unlock( &mutex );

        RayTraceRenderer *rtr = ((RayTraceRenderer*)rayTraceRenderer);
        rtr->Traverser::preprocess(myThreadID);
        pthread_mutex_lock( &mutex );
        runningCounter--;
        pthread_cond_signal(&condition_end);
    }
    return NULL;
}
#endif //NUM_RENDER_THREADS

void RayTraceRenderer::render(bool update)
{
  Clock preprocessClock; preprocessClock.start();

if  (update) {

#ifndef TRI_PREPROCESS
  TriangleIntersector::preprocess(); //place it in front only for debugging
  //and measuring just the parallel build timings. This makes the end code
  //slower
#endif //TRI_PREPROCESS

  runningCounter = NUM_BUILD_THREADS;

  pthread_mutex_lock( &mutex );
  pthread_cond_broadcast(&condition_preprocess_start); //wake up preprocess threads
  pthread_mutex_unlock( &mutex );

#ifndef TRI_PREPROCESS
//   TriangleIntersector::preprocess();
#endif //TRI_PREPROCESS

  //wait for threads to finish.
  while(runningCounter) {
      pthread_mutex_lock( &mutex );
      if (runningCounter==0) {
          pthread_mutex_unlock( &mutex ); 
          break;
      }
      pthread_cond_wait( &condition_end, &mutex );
      pthread_mutex_unlock( &mutex );
  }

#else
  TriangleIntersector::preprocess();
  Traverser::preprocess();
#endif //NUM_TRI_INTERSECT_THREADS && NUM_BUILD_THREADS
}
 
 preprocessClock.stop();
 // cout << "preprocess time: " << preprocessClock.msecs() <<"ms" <<endl;
}

